open Linear_types
open Gmp

type lp_result =
  | RESULT_OPTIMAL of Q.t * (string, Q.t * (Q.t option) * (Q.t option)) Hashtbl.t
  | RESULT_UNBOUNDED
  | RESULT_INFEASIBLE
  | RESULT_DUAL_INFEASIBLE
  | RESULT_UNSUPPORTED;;

type wide_constraint_kind =
  | LESS_EQ
  | GREATER_EQ
  | EQUAL;;

type 'variable problem = {
    constraints : (('variable, wide_constraint_kind) lconstraint) list;
    bounds : ('variable * (Q.t option) * (Q.t option)) list; 
    objective : 'variable linear_form;
    direction : optimization_mode };;

let print_constraint (form, kind, rhs) =
  print_form form;
  Printf.printf " %s %s\n"
    (match kind with
    | LESS_EQ -> "<="
    | GREATER_EQ -> ">="
    | EQUAL -> "=")
    (Q.to_string rhs);;
  
let print_bound (var_name, low, high) =
  Printf.printf "%s <= %s <= %s\n"
    (match low with
    | None -> "-inf"
    | Some x -> Q.to_string x)
    var_name
    (match high with
    | None -> "+inf"
    | Some x -> Q.to_string x);;

let print_problem problem =
  Printf.printf "LP problem:\n";
  List.iter print_constraint problem.constraints;
  List.iter print_bound problem.bounds;
  Printf.printf "%s objective: "
    (match problem.direction with
    | MINIMIZATION -> "minimize"
    | MAXIMIZATION -> "maximize");
  print_form problem.objective;
  print_newline ();;

let rec list_map f = function
  | [] -> []
  | h::t ->

      let fst = f h in
      fst::(list_map f t);;
(*
let insert_coeffs lp var_hash linear_form row =
 List.iter (fun (var_name, coeff) ->
      Exlp.lp_set_coefficient lp coeff row
     (Hashtbl.find var_hash var_name)) linear_form;;

let exlp_collect_variables lp var_hash linear_form =
    List.iter (fun (var_name, coeff) ->
      if not (Hashtbl.mem var_hash var_name)
      then
	let id = Exlp.lp_add_var_without_a lp var_name in
	Hashtbl.add var_hash var_name id ) linear_form;;

let exlp_solve problem =
  try
    let lp = Exlp.lp_create()
    and var_hash = Hashtbl.create 100 in
    Exlp.lp_set_verbose lp false;
    
    let cur_row = ref 1 in
    let rows = list_map
	(fun (form, kind, rhs) ->
	  let row = Exlp.lp_add_row lp (Printf.sprintf "row%d" !cur_row) in
	  Exlp.lp_set_row_equality lp row
	    (match kind with
	    | LESS_EQ -> 'L'
	    | GREATER_EQ -> 'G'
	    | EQUAL -> 'E');
	  incr cur_row;
	  row) problem.constraints in
    Exlp.lp_set_obj_name lp "obj";
    let obj = Exlp.lp_get_row_num lp "obj" in

    exlp_collect_variables lp var_hash problem.objective;
    List.iter (fun (form, kind, rhs) ->
      exlp_collect_variables lp var_hash form) problem.constraints;
    
    Exlp.lp_matrix_adjust_size lp;
    List.iter2 (fun (form, kind, rhs) row ->
      insert_coeffs lp var_hash form row) problem.constraints rows;
    insert_coeffs lp var_hash problem.objective obj;
    
    Exlp.lp_set_maximize lp
      (match problem.direction with MINIMIZATION->false | MAXIMIZATION->true);
    
    List.iter2 (fun (form, kind, rhs) row ->
      Exlp.lp_set_rhs lp row rhs) problem.constraints rows;

    (let var_bounds = Array.create (Exlp.lp_get_vars lp) (None, None) in
    List.iter (fun (var, lower, upper) ->
      try var_bounds.(Hashtbl.find var_hash var) <- (lower, upper)
      with Not_found -> ()) problem.bounds;
    Array.iteri (fun i (lower, upper) ->
      Exlp.lp_set_lower_bound lp i lower;
      Exlp.lp_set_upper_bound lp i upper) var_bounds);

    match Exlp.solve_lp lp with
    | Exlp.LP_RESULT_OPTIMAL -> RESULT_OPTIMAL
	  ((Exlp.lp_get_object_value lp),
      (let out_hash = Hashtbl.create 100 in
      Hashtbl.iter (fun var_name var_id ->
	Hashtbl.add out_hash var_name
	  ((Exlp.lp_get_solution lp var_id),
	  (Exlp.lp_get_lower_bound lp var_id),
	  (Exlp.lp_get_upper_bound lp var_id)))
	var_hash;
      out_hash))
    | Exlp.LP_RESULT_UNBOUNDED -> RESULT_UNBOUNDED
    | Exlp.LP_RESULT_INFEASIBLE -> RESULT_INFEASIBLE
    | Exlp.LP_RESULT_DUAL_INFEASIBLE -> RESULT_DUAL_INFEASIBLE
    | Exlp.LP_RESULT_UNSUPPORTED -> RESULT_UNSUPPORTED
  with Not_found -> failwith "bad handling of variables";;
*)

let ppl_solve problem =
  let var_hash = Hashtbl.create 100
  and var_index = ref 0 in
  let collect_variable var_name =
    if not (Hashtbl.mem var_hash var_name)
    then
      let id = !var_index in
	Hashtbl.add var_hash var_name id;
        var_index := succ id in
  let collect_variables_in_form linear_form =
    List.iter (fun (var_name, coeff) -> collect_variable var_name) linear_form
  and map_variable v = Hashtbl.find var_hash v 
  in
    collect_variables_in_form problem.objective;
    List.iter (fun (form, kind, rhs) ->
		 collect_variables_in_form form) problem.constraints;
    List.iter (fun (var_name, _, _) ->
		 collect_variable var_name) problem.bounds;
    let constraints =
      List.fold_right
	(fun (var_name, low, high) already ->
	   let var_id = map_variable var_name in
	   let already1 =
	     match low with
	       | None -> already
	       | Some x ->
		   (PPL.constraint_make
		      (PPL.linear_expr_from_scaled_variable
		       (Q.get_den x) var_id)
		      Linear_types.GREATER_EQ
		    (PPL.linear_expr_from_z (Q.get_num x)))::already in
	     match high with
	       | None -> already1
	       | Some x ->
		   (PPL.constraint_make
		      (PPL.linear_expr_from_scaled_variable
		       (Q.get_den x) var_id)
		      Linear_types.LESS_EQ
		    (PPL.linear_expr_from_z (Q.get_num x)))::already1
	) problem.bounds
      (List.map
	 (fun c ->
	    let (lhs, kind, rhs) =
	      map_variables_in_constraint map_variable
		(zlconstraint_of_lconstraint c) in
	      PPL.constraint_make (PPL.linear_expr_make lhs)
		(match kind with LESS_EQ -> Linear_types.LESS_EQ
		| GREATER_EQ -> Linear_types.GREATER_EQ
		| EQUAL -> Linear_types.EQUAL)
		(PPL.linear_expr_from_z rhs))
	 problem.constraints)
    and objective_factor =List.fold_right (fun (name, coeff) z ->
      Z.lcm (Q.get_den coeff) z) problem.objective Z.one in
    let objective = PPL.linear_expr_make (map_variables_in_form map_variable
      (zlform_of_scaled_lform objective_factor problem.objective)) in
    let ppl_problem = PPL.lp_problem_create constraints objective problem.direction in
    match PPL.lp_problem_solve ppl_problem with
      | PPL.UNFEASIBLE_LP_PROBLEM -> RESULT_INFEASIBLE
      | PPL.UNBOUNDED_LP_PROBLEM -> RESULT_UNBOUNDED
      | PPL.OPTIMIZED_LP_PROBLEM ->
	  let optimal_value =
	    let (num, den) = PPL.lp_problem_optimal_value ppl_problem in
	    Q.div (Q.from_z num) (Q.from_z (Z.mul den objective_factor))
          and generator = PPL.lp_problem_optimizing_point ppl_problem in
	  let generator_divisor = PPL.generator_read_divisor generator
	  and generator_coeffs = PPL.generator_read_coefficients generator in
	  let generator_array = Array.create !var_index Z.zero in
          List.iter (fun (i, z) -> generator_array.(i)<-z) generator_coeffs;
          let out_hash = Hashtbl.create 100 in
	    Hashtbl.iter
	      (fun var_name var_id ->
		 Hashtbl.add out_hash var_name
		   ((Q.div (Q.from_z generator_array.(var_id))
		      (Q.from_z generator_divisor)), None, None)) var_hash;
	  RESULT_OPTIMAL(optimal_value, out_hash);;

let solve problem = ppl_solve problem;;
