#undef DEBUG

#ifdef DEBUG
#include <stdio.h>
#endif

#include <gmp.h>

#include "lpstruct.h"
#include "mylib.h"
#include "solve_lp.h"

#define MODULE "Exlp"

#include "ml_interface.h"

#define CAMLCALL(id) value ml_exlp_##id

static inline mpq_t * mpq_val (value val)
{
  return ((mpq_t *) (Data_custom_val(val)));
}

static inline LP* lp_of_value(value r) {
  return *((LP**) Data_custom_val(r));
}

void ml_exlp_lp_finalize(value r) {
  lp_free(lp_of_value(r));
}

static struct custom_operations ml_exlp_lp_custom = {
  "Exlp.lp",
  &ml_exlp_lp_finalize
};


CAMLCALL(lp_create)(value dummy) {
  CAMLparam0();
  CAMLlocal1(r);
  LP *lp = new_lp(NULL);
  r = caml_alloc_custom(&ml_exlp_lp_custom,
		   sizeof(LP*),
		   0,
		   1);
  *((LP**) Data_custom_val(r)) = lp;
  /* If print_sol is not set to 1, then the output vector is incorrect
     (some preprocessing has destroyed it) */
#ifdef DEBUG
  lp->verbose = 1;
#else
  lp->verbose = 0;
#endif
  lp->print_sol = 1;
  lp->scaling = 0;
  lp_hash_str_init(lp, lp->hash_entries);
#ifndef NO_GMP_HASH
  my_hash_mpq_init(lp->hash_entries);
#endif

  CAMLreturn(r);
}

CAMLCALL(lp_set_verbose)(value lp, value flag) {
  CAMLparam2(lp, flag);
  lp_of_value(lp)->verbose = Bool_val(flag);
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_set_obj_name)(value lp, value name) {
  CAMLparam2(lp, name);
  lp_set_obj_name(lp_of_value(lp), String_val(name));
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_add_row)(value lp, value name) {
  CAMLparam2(lp, name);
  CAMLreturn(Val_int(lp_add_row(lp_of_value(lp), String_val(name))));
}

CAMLCALL(lp_get_row_num)(value lp, value name) {
  CAMLparam2(lp, name);
  CAMLreturn(Val_int(lp_get_row_num(lp_of_value(lp), String_val(name))));
}

CAMLCALL(lp_set_row_equality)(value lp, value row, value kind) {
  CAMLparam3(lp, row, kind);
  lp_set_row_equality(lp_of_value(lp), Int_val(row), Int_val(kind));
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_add_var_without_a)(value lp, value name) {
  CAMLparam2(lp, name);
  CAMLreturn(Val_int(lp_add_var_without_A(lp_of_value(lp), String_val(name))));
}

CAMLCALL(lp_get_var_num)(value lp, value name) {
  CAMLparam2(lp, name);
  CAMLreturn(Val_int(lp_get_var_num(lp_of_value(lp), String_val(name))));
}

CAMLCALL(lp_get_vars)(value lp) {
  CAMLparam1(lp);
  CAMLreturn(Val_int(lp_of_value(lp)->vars));
}

CAMLCALL(lp_matrix_adjust_size)(value val_lp) {
  CAMLparam1(val_lp);
  LP *lp = lp_of_value(val_lp);
#ifdef DEBUG
  printf("%d x %d problem\n", lp->rows, lp->vars);
#endif
  matrix_resize(lp->A, lp->rows, lp->vars);
  vector_resize(lp->b, lp->rows);
  vector_resize(lp->xb, lp->rows);
  vector_resize(lp->cb, lp->rows);  
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_set_coefficient)(value lp, value coeff, value row, value var) {
  CAMLparam4(lp, coeff, row, var);
  lp_set_coefficient(lp_of_value(lp), *mpq_val(coeff), Int_val(row), Int_val(var));
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_set_rhs)(value lp, value row, value rhs) {
  CAMLparam3(lp, row, rhs);
  lp_set_rhs(lp_of_value(lp), Int_val(row), *mpq_val(rhs));
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_set_maximize)(value vlp, value flag) {
  CAMLparam2(vlp, flag);
  LP *lp = lp_of_value(vlp);
  if ((lp ->maximize = Bool_val(flag)) == FALSE)
    vector_rev_sgn(lp->c);
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_get_object_value2)(value lp, value ret) {
  CAMLparam2(lp, ret);
  lp_get_object_value(lp_of_value(lp), mpq_val(ret));
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_print_basis)(value lp) {
  CAMLparam1(lp);
  lp_print_basis(lp_of_value(lp));
  CAMLreturn(Val_unit);
}

CAMLCALL(solve_lp)(value vlp) {
  CAMLparam1(vlp);
  LP *lp = lp_of_value(vlp);
  int ret = solve_lp(lp);
#ifdef DEBUG
  printf("solver returns %d\n", ret);
#endif
  CAMLreturn(Val_int(ret));
}

CAMLCALL(lp_get_lower_bound_is_valid)(value lp, value var) {
  CAMLparam2(lp, var);
  CAMLreturn(Val_bool(lp_of_value(lp)->lower.is_valid[Int_val(var)]));
}

CAMLCALL(lp_get_upper_bound_is_valid)(value lp, value var) {
  CAMLparam2(lp, var);
  CAMLreturn(Val_bool(lp_of_value(lp)->upper.is_valid[Int_val(var)]));
}

CAMLCALL(lp_set_lower_bound_is_valid)(value lp, value var, value flag) {
  CAMLparam3(lp, var, flag);
  if (lp_of_value(lp)->lower.is_valid[Int_val(var)] && !Bool_val(flag)) {
    mpq_clear(lp_of_value(lp)->lower.bound[Int_val(var)]);
  }
  else if (!lp_of_value(lp)->lower.is_valid[Int_val(var)] && Bool_val(flag)) {
    mpq_init(lp_of_value(lp)->lower.bound[Int_val(var)]);
  }

  lp_of_value(lp)->lower.is_valid[Int_val(var)] = Bool_val(flag);
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_set_upper_bound_is_valid)(value lp, value var, value flag) {
  CAMLparam3(lp, var, flag);
  if (lp_of_value(lp)->upper.is_valid[Int_val(var)] && !Bool_val(flag)) {
    mpq_clear(lp_of_value(lp)->upper.bound[Int_val(var)]);
  }
  else if (!lp_of_value(lp)->upper.is_valid[Int_val(var)] && Bool_val(flag)) {
    mpq_init(lp_of_value(lp)->upper.bound[Int_val(var)]);
  }

  lp_of_value(lp)->upper.is_valid[Int_val(var)] = Bool_val(flag);
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_get_lower_bound2)(value lp, value var, value ret) {
  CAMLparam3(lp, var, ret);
  mpq_set(*mpq_val(ret), lp_of_value(lp)->lower.bound[Int_val(var)]);
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_get_upper_bound2)(value lp, value var, value ret) {
  CAMLparam3(lp, var, ret);
  mpq_set(*mpq_val(ret), lp_of_value(lp)->upper.bound[Int_val(var)]);
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_set_lower_bound2)(value lp, value var, value bound) {
  CAMLparam3(lp, var, bound);
  mpq_set(lp_of_value(lp)->lower.bound[Int_val(var)], *mpq_val(bound));
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_set_upper_bound2)(value lp, value var, value bound) {
  CAMLparam3(lp, var, bound);
  mpq_set(lp_of_value(lp)->upper.bound[Int_val(var)], *mpq_val(bound));
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_print_nonzero_vars)(value lp) {
  CAMLparam1(lp);
  lp_print_nonzero_vars(lp_of_value(lp));
  CAMLreturn(Val_unit);
}

CAMLCALL(lp_get_solution2)(value lp, value var, value ret) {
  CAMLparam3(lp, var, ret);
  mpq_set(*mpq_val(ret), *vector_get_element_ptr(lp_of_value(lp)->x, Int_val(var)));
  CAMLreturn(Val_unit);
}

CAMLCALL(mylib_init)(value dummy) {
  CAMLparam0();
#ifdef DEBUG
  printf("initializing library\n");
#endif
  mylib_init();
  CAMLreturn(Val_unit);
}
