open Gmp

let float_rounding = true

let rec tuple_map
    (op : 'a Proposition.proposition -> 'a Proposition.proposition
      -> 'a Proposition.proposition)
    (neutral : 'a Proposition.proposition)
 = function
  | [] -> neutral
  | [x] -> x
  | h::t -> op h (tuple_map op neutral t);;

let vand l = tuple_map (fun x y -> Proposition.And(x, y)) Proposition.True l
let vor l = tuple_map (fun x y -> Proposition.Or(x, y)) Proposition.False l
let vnot x = Proposition.Not(x)
let implies p1 p2 = Proposition.Implies(p1, p2)

let less x y = Proposition.Atomic(Quantified_formula.QAtomic(Atomic_formula.Syntax_comparison(x, Atomic_formula.LESS, y)))
let greater x y = Proposition.Atomic(Quantified_formula.QAtomic(Atomic_formula.Syntax_comparison(x, Atomic_formula.GREATER, y)))
let less_eq x y = Proposition.Atomic(Quantified_formula.QAtomic(Atomic_formula.Syntax_comparison(x, Atomic_formula.LESS_EQ, y)))
let greater_eq x y = Proposition.Atomic(Quantified_formula.QAtomic(Atomic_formula.Syntax_comparison(x, Atomic_formula.GREATER_EQ, y)))
let equal x y =  Proposition.Atomic(Quantified_formula.QAtomic(Atomic_formula.Syntax_comparison(x, Atomic_formula.EQUAL, y)))

let const x : Atomic_formula.linear_expression = ([], x)
let var x : Atomic_formula.linear_expression = ([x, Q.one], Q.zero)
let plus x y = Atomic_formula.add_expressions x y
let minus x y = Atomic_formula.sub_expressions x y
let scale s x = Atomic_formula.scale_expression s x
let neg x = Atomic_formula.neg_expression x

let forall vars x = Proposition.Atomic(Quantified_formula.Forall(vars, x))
let exists vars x = Proposition.Atomic(Quantified_formula.Exists(vars, x))

let ifte c e1 e2 = vor [ vand [c; e1]; vand [vnot c; e2]];;

let compile f =
  Quantified_formula.compile_qformula Quantified_formula.compile_atomic_q f;;

let fresh_variable =
  let table = Hashtbl.create 100 in
  function name ->
    let count =
      try Hashtbl.find table name
      with Not_found -> 0 in
    Hashtbl.replace table name (count + 1);
    Printf.sprintf "%s$%d" name count;;

let syntax_subst = Quantified_formula.syntax_qsubstitute;;
let pr_subst = Quantified_formula.pr_qsubstitute;;

let pr_is_minimum v prop =
  let v' = fresh_variable v in
  forall [v'] (vand [ prop ;
		       (implies (pr_subst v (var v') prop) (compile (less_eq (var v) (var v')))) ]);;

let syntax_is_minimum v prop =
  let v' = fresh_variable v in
  forall [v'] (vand [ prop ;
		      (implies (syntax_subst v (var v') prop) (less_eq (var v) (var v'))) ]);;

let pr_is_maximum v prop =
  vand [ prop ;
	 let v' = fresh_variable v in
	 forall [v'] (implies (pr_subst v (var v') prop) (compile (greater_eq (var v) (var v')))) ];;

let syntax_is_maximum v prop =
  vand [ prop ;
	 let v' = fresh_variable v in
	 forall [v'] (implies (syntax_subst v (var v') prop) (greater_eq (var v) (var v'))) ];;

let eps_rel = Q.div Q.one (Q.from_z (Z.mul2exp Z.one 53))
let eps_abs = Q.div Q.one (Q.from_z (Z.mul2exp Z.one 1023))

let is_rounded rounded unrounded =
  if float_rounding
  then
    vor [vand [less_eq (var rounded) (const eps_abs);
	       greater_eq (var rounded) (const Q.zero);
	       less_eq unrounded (const eps_abs);
	       greater_eq unrounded (const Q.zero)];
	 
	 vand [greater_eq (var rounded) (const (Q.neg eps_abs));
	       less_eq (var rounded) (const Q.zero);
	       greater_eq unrounded (const (Q.neg eps_abs));
	       less_eq unrounded (const Q.zero)];
	 
       vand [greater_eq (var rounded) (const eps_abs);
	     greater_eq unrounded (const eps_abs);
	     less_eq (var rounded) (scale (Q.add Q.one eps_rel) unrounded);
	     greater_eq (var rounded) (scale (Q.sub Q.one eps_rel) unrounded)];
       
       vand [less_eq (var rounded) (const (Q.neg eps_abs));
	     less_eq unrounded (const (Q.neg eps_abs));
	     greater_eq (var rounded) (scale (Q.add Q.one eps_rel) unrounded);
	     less_eq (var rounded) (scale (Q.sub Q.one eps_rel) unrounded)]]
  else equal (var rounded) unrounded;;

let postcondition =
  vand [ less_eq (var "s1") (var "s1max");
	 greater_eq (var "s1") (var "s1min") ];;

let precondition =
   vand [ less_eq (var "e1") (var "e1max");
	  greater_eq (var "e1") (var "e1min");
	  less_eq (var "e2") (var "e2max");
	  greater_eq (var "e2") (var "e2min");
	  less_eq (var "e3") (var "e3max");
	  greater_eq (var "e3") (var "e3min");
	  less_eq (var "olds1") (var "s1max");
	  greater_eq (var "olds1") (var "s1min") ];;

let step1 =
  vand [ is_rounded "delta1" (minus (var "e1") (var "olds1"));
	 ifte (less (var "delta1") (var "e2")) (equal (var "o1") (var "e1"))
	     (equal (var "o1") (plus (var "olds1") (var "e2"))) ];;

let step2 =
  vand [ is_rounded "delta2" (minus (var "o1") (var "olds1"));
	 ifte (greater (var "delta2") (neg (var "e2")))
	     (equal (var "o2") (var "o1"))
	     (equal (var "o2") (minus (var "o1") (var "e2"))) ];;

let step3 = vor [ equal (var "s1") (var "o2");
		  equal (var "s1") (var "e3") ];;

let correct =
  vand [ less_eq (var "e1min") (var "e1max");
	 less_eq (var "e2min") (var "e2max");
	 less_eq (var "e3min") (var "e3max");
	 less_eq (var "s1min") (var "s1max");
	 greater (var "e2min") (const Q.zero) ];;

let stable =
  forall [ "olds1"; "e1"; "e2"; "e3"; "o1"; "o2"; "delta1"; "delta2"; "s1" ]
    (vand [correct;
	   implies ( vand [ precondition; step1; step2; step3 ] )
	     postcondition ]);;

let stable_s1max = forall [] (exists [ "s1min" ] stable);;

let do_substs =
  List.fold_right
     (fun (var, value) formula -> pr_subst var value formula)

let instances1 =
  [(* "e1min", (const (Q.from_int (-10000)));
    "e1max", (const (Q.from_int (10000)));
    "e2min", (const (Q.from_int 1));
    "e2max", (const (Q.from_int 3));
    "e3min", (const (Q.from_int (-5000)));
    "e3max", (const (Q.from_int (5000))) *) ];;

print_endline "Stable sets:";;
let elim_stable = Proposition.simplify (Qelim.qelim (do_substs instances1 (compile stable_s1max)));;
Atomic_formula.print_pr_formula stdout elim_stable;;
print_newline();;
Gc.full_major();;

print_endline "Above some stable set:";;
let above =
  let v' = fresh_variable "s1max" in
  exists [ v' ]
    (vand [compile (greater_eq (var "s1max") (var v'));
	   pr_subst "s1max" (var v') (Quantified_formula.lift elim_stable)]);;

let elim_above =  Proposition.simplify (Qelim.qelim above);;
Atomic_formula.print_pr_formula stdout elim_above;;
print_newline();;
Gc.full_major();;

print_endline "Strictly above some stable set:";;
let strictly_above =
  let v' = fresh_variable "s1max" in
  exists [ v' ]
    (vand [compile (greater (var "s1max") (var v'));
	   pr_subst "s1max" (var v') (Quantified_formula.lift elim_above)]);;

let elim_strictly_above =  Proposition.simplify (Qelim.qelim strictly_above);;
Atomic_formula.print_pr_formula stdout elim_strictly_above;;
print_newline();;
Gc.full_major();;


print_endline "Minimal sets:";;
let minimal = vand [elim_above ; vnot elim_strictly_above];;
let elim_minimal = Proposition.simplify (Qelim.qelim (exists [] (Quantified_formula.lift minimal)));;
Atomic_formula.print_pr_formula stdout elim_minimal;;
print_newline();;
Gc.full_major();;

let instances2 =
  [ "e1min", (const (Q.from_int (-10000)));
    "e1max", (const (Q.from_int (10000)));
    "e2min", (const (Q.from_int 1));
    "e2max", (const (Q.from_int 3));
    "e3min", (const (Q.from_int (-5000)));
    "e3max", (const (Q.from_int (5000))) ];;

print_endline "Instanciation:";;
let instanciated = Qelim.qelim
        (do_substs instances2
	  (Quantified_formula.lift elim_minimal));;

let instanciated2 = Proposition.simplify (Qelim.qelim (forall [] (exists [] (Quantified_formula.lift instanciated))));;

Atomic_formula.print_pr_formula stdout instanciated2;;
print_newline();;
